@@ -5,6 +5,7 @@ import asyncio
 from collections import defaultdict
 import functools
 import logging
+import secrets
 from typing import (
     Any,
     Callable,
@@ -17,6 +18,7 @@ from typing import (
     Sequence,
     Type,
     TypeVar,
+    TypedDict,
     Union,
     cast,
 )
@@ -24,11 +26,16 @@ import uuid
 import weakref
 
 import aioredis
+import aioredis.exceptions
+import aioredis.sentinel
+from aiotools.context import aclosing
 import attr
 
 from . import msgpack, redis
 from .logging import BraceStyleAdapter
 from .types import (
+    EtcdRedisConfig,
+    RedisConnectionInfo,
     aobject,
     AgentId,
     KernelId,
@@ -83,6 +90,10 @@ class DoScheduleEvent(EmptyEventArgs, AbstractEvent):
     name = "do_schedule"
 
 
+class DoPrepareEvent(EmptyEventArgs, AbstractEvent):
+    name = "do_prepare"
+
+
 class DoIdleCheckEvent(EmptyEventArgs, AbstractEvent):
     name = "do_idle_check"
 
@@ -145,7 +156,7 @@ class AgentErrorEvent(AbstractEvent):
             self.traceback,
             self.user,
             self.context_env,
-            str(self.severity),
+            self.severity.value,
         )
 
     @classmethod
@@ -215,6 +226,10 @@ class KernelStartedEvent(KernelCreationEventArgs, AbstractEvent):
     name = "kernel_started"
 
 
+class KernelCancelledEvent(KernelCreationEventArgs, AbstractEvent):
+    name = "kernel_cancelled"
+
+
 @attr.s(slots=True, frozen=True)
 class KernelTerminationEventArgs():
     kernel_id: KernelId = attr.ib()
@@ -237,10 +252,6 @@ class KernelTerminationEventArgs():
         )
 
 
-class KernelCancelledEvent(KernelTerminationEventArgs, AbstractEvent):
-    name = "kernel_cancelled"
-
-
 class KernelTerminatingEvent(KernelTerminationEventArgs, AbstractEvent):
     name = "kernel_terminating"
 
@@ -279,6 +290,10 @@ class SessionScheduledEvent(SessionCreationEventArgs, AbstractEvent):
     name = "session_scheduled"
 
 
+class SessionPreparingEvent(SessionCreationEventArgs, AbstractEvent):
+    name = "session_preparing"
+
+
 class SessionCancelledEvent(SessionCreationEventArgs, AbstractEvent):
     name = "session_cancelled"
 
@@ -472,9 +487,9 @@ class BgtaskFailedEvent(BgtaskDoneEventArgs, AbstractEvent):
 
 
 class RedisConnectorFunc(Protocol):
-    async def __call__(
+    def __call__(
         self,
-    ) -> aioredis.abc.AbcPool:
+    ) -> aioredis.ConnectionPool:
         ...
 
 
@@ -491,8 +506,71 @@ EventCallback = Union[
 @attr.s(auto_attribs=True, slots=True, frozen=True, eq=False, order=False)
 class EventHandler(Generic[TContext, TEvent]):
     event_cls: Type[TEvent]
+    name: str
     context: TContext
     callback: EventCallback[TContext, TEvent]
+    coalescing_opts: Optional[CoalescingOptions]
+    coalescing_state: CoalescingState
+
+
+class CoalescingOptions(TypedDict):
+    max_wait: float
+    max_batch_size: int
+
+
+@attr.s(auto_attribs=True, slots=True)
+class CoalescingState:
+    batch_size: int = 0
+    last_added: float = 0.0
+    last_handle: asyncio.TimerHandle | None = None
+    fut_sync: asyncio.Future | None = None
+
+    def proceed(self):
+        if self.fut_sync is not None and not self.fut_sync.done():
+            self.fut_sync.set_result(None)
+
+    async def rate_control(self, opts: CoalescingOptions | None) -> bool:
+        if opts is None:
+            return True
+        loop = asyncio.get_running_loop()
+        if self.fut_sync is None:
+            self.fut_sync = loop.create_future()
+        assert self.fut_sync is not None
+        self.last_added = loop.time()
+        self.batch_size += 1
+        if self.batch_size >= opts['max_batch_size']:
+            assert self.last_handle is not None
+            self.last_handle.cancel()
+            self.fut_sync.cancel()
+            self.last_handle = None
+            self.last_added = 0.0
+            self.batch_size = 0
+            return True
+        # Schedule.
+        self.last_handle = loop.call_later(
+            opts['max_wait'],
+            self.proceed,
+        )
+        if self.last_added > 0 and loop.time() - self.last_added < opts['max_wait']:
+            # Cancel the previously pending task.
+            self.last_handle.cancel()
+            self.fut_sync.cancel()
+            # Reschedule.
+            self.fut_sync = loop.create_future()
+            self.last_handle = loop.call_later(
+                opts['max_wait'],
+                self.proceed,
+            )
+        try:
+            await self.fut_sync
+        except asyncio.CancelledError:
+            return False
+        else:
+            self.fut_sync = None
+            self.last_handle = None
+            self.last_added = 0.0
+            self.batch_size = 0
+            return True
 
 
 class EventDispatcher(aobject):
@@ -512,25 +590,31 @@ class EventDispatcher(aobject):
 
     consumers: defaultdict[str, set[EventHandler[Any, AbstractEvent]]]
     subscribers: defaultdict[str, set[EventHandler[Any, AbstractEvent]]]
-    redis_consumer: aioredis.Redis
-    redis_subscriber: aioredis.Redis
+    redis_client: RedisConnectionInfo
     consumer_loop_task: asyncio.Task
     subscriber_loop_task: asyncio.Task
     consumer_taskset: weakref.WeakSet[asyncio.Task]
     subscriber_taskset: weakref.WeakSet[asyncio.Task]
 
-    _connector: RedisConnectorFunc
     _log_events: bool
+    _consumer_name: str
 
-    def __init__(self, connector: RedisConnectorFunc, log_events: bool = False) -> None:
-        self._connector = connector
+    def __init__(
+        self,
+        connector: EtcdRedisConfig,
+        db: int = 0,
+        log_events: bool = False,
+        *,
+        consumer_group: str = "manager",
+    ) -> None:
+        self.redis_client = redis.get_redis_object(connector, db=db)
         self._log_events = log_events
         self.consumers = defaultdict(set)
         self.subscribers = defaultdict(set)
+        self._consumer_group = consumer_group
+        self._consumer_name = secrets.token_urlsafe(16)
 
     async def __ainit__(self) -> None:
-        self.redis_consumer = await self._connector()
-        self.redis_subscriber = await self._connector()
         self.consumer_loop_task = asyncio.create_task(self._consume_loop())
         self.subscriber_loop_task = asyncio.create_task(self._subscribe_loop())
         self.consumer_taskset = weakref.WeakSet()
@@ -551,18 +635,20 @@ class EventDispatcher(aobject):
         cancelled_tasks.append(self.consumer_loop_task)
         cancelled_tasks.append(self.subscriber_loop_task)
         await asyncio.gather(*cancelled_tasks, return_exceptions=True)
-        self.redis_consumer.close()
-        self.redis_subscriber.close()
-        await self.redis_consumer.wait_closed()
-        await self.redis_subscriber.wait_closed()
+        await self.redis_client.close()
 
     def consume(
         self,
         event_cls: Type[TEvent],
         context: TContext,
         callback: EventCallback[TContext, TEvent],
+        coalescing_opts: CoalescingOptions = None,
+        *,
+        name: str = None,
     ) -> EventHandler[TContext, TEvent]:
-        handler = EventHandler(event_cls, context, callback)
+        if name is None:
+            name = f"evh-{secrets.token_urlsafe(16)}"
+        handler = EventHandler(event_cls, name, context, callback, coalescing_opts, CoalescingState())
         self.consumers[event_cls.name].add(cast(EventHandler[Any, AbstractEvent], handler))
         return handler
 
@@ -577,8 +663,13 @@ class EventDispatcher(aobject):
         event_cls: Type[TEvent],
         context: TContext,
         callback: EventCallback[TContext, TEvent],
+        coalescing_opts: CoalescingOptions = None,
+        *,
+        name: str = None,
     ) -> EventHandler[TContext, TEvent]:
-        handler = EventHandler(event_cls, context, callback)
+        if name is None:
+            name = f"evh-{secrets.token_urlsafe(16)}"
+        handler = EventHandler(event_cls, name, context, callback, coalescing_opts, CoalescingState())
         self.subscribers[event_cls.name].add(cast(EventHandler[Any, AbstractEvent], handler))
         return handler
 
@@ -588,6 +679,24 @@ class EventDispatcher(aobject):
     ) -> None:
         self.subscribers[handler.event_cls.name].discard(cast(EventHandler[Any, AbstractEvent], handler))
 
+    async def handle(self, evh_type: str, evh: EventHandler, source: AgentId, args: tuple) -> None:
+        loop = asyncio.get_running_loop()
+        coalescing_opts = evh.coalescing_opts
+        coalescing_state = evh.coalescing_state
+        cb = evh.callback
+        event_cls = evh.event_cls
+        if (await coalescing_state.rate_control(coalescing_opts)):
+            if self._log_events:
+                log.debug("DISPATCH_{}(evh:{})", evh_type, evh.name)
+            if asyncio.iscoroutinefunction(cb):
+                # mypy cannot catch the meaning of asyncio.iscoroutinefunction().
+                await cb(evh.context, source, event_cls.deserialize(args))  # type: ignore
+            else:
+                wrapped_cb = functools.partial(
+                    cb, evh.context, source, event_cls.deserialize(args),  # type: ignore
+                )
+                loop.call_soon(wrapped_cb)
+
     async def dispatch_consumers(
         self,
         event_name: str,
@@ -595,23 +704,12 @@ class EventDispatcher(aobject):
         args: tuple,
     ) -> None:
         if self._log_events:
-            log_fmt = 'DISPATCH_CONSUMERS(ev:{}, ag:{})'
-            log_args = (event_name, source)
-            log.debug(log_fmt, *log_args)
-        loop = asyncio.get_running_loop()
-        for consumer in self.consumers[event_name]:
-            cb = consumer.callback
-            event_cls = consumer.event_cls
-            if asyncio.iscoroutinefunction(cb):
-                # mypy cannot catch the meaning of asyncio.iscoroutinefunction().
-                self.consumer_taskset.add(asyncio.create_task(
-                    cb(consumer.context, source, event_cls.deserialize(args))  # type: ignore
-                ))
-            else:
-                cb = functools.partial(
-                    cb, consumer.context, source, event_cls.deserialize(args),  # type: ignore
-                )
-                loop.call_soon(cb)
+            log.debug('DISPATCH_CONSUMERS(ev:{}, ag:{})', event_name, source)
+        for consumer in self.consumers[event_name].copy():
+            self.consumer_taskset.add(asyncio.create_task(
+                self.handle("CONSUMER", consumer, source, args)
+            ))
+            await asyncio.sleep(0)
 
     async def dispatch_subscribers(
         self,
@@ -620,75 +718,72 @@ class EventDispatcher(aobject):
         args: tuple,
     ) -> None:
         if self._log_events:
-            log_fmt = 'DISPATCH_SUBSCRIBERS(ev:{}, ag:{})'
-            log_args = (event_name, source)
-            log.debug(log_fmt, *log_args)
-        loop = asyncio.get_running_loop()
-        for subscriber in self.subscribers[event_name]:
-            cb = subscriber.callback
-            event_cls = subscriber.event_cls
-            if asyncio.iscoroutinefunction(cb):
-                # mypy cannot catch the meaning of asyncio.iscoroutinefunction().
-                self.subscriber_taskset.add(asyncio.create_task(
-                    cb(subscriber.context, source, event_cls.deserialize(args))  # type: ignore
-                ))
-            else:
-                cb = functools.partial(
-                    cb, subscriber.context, source, event_cls.deserialize(args),  # type: ignore
-                )
-                loop.call_soon(cb)
+            log.debug('DISPATCH_SUBSCRIBERS(ev:{}, ag:{})', event_name, source)
+        for subscriber in self.subscribers[event_name].copy():
+            self.subscriber_taskset.add(asyncio.create_task(
+                self.handle("SUBSCRIBER", subscriber, source, args)
+            ))
+            await asyncio.sleep(0)
 
     async def _consume_loop(self) -> None:
-        while True:
-            try:
-                key, raw_msg = await redis.execute_with_retries(
-                    lambda: self.redis_consumer.blpop('events.prodcons'))
-                msg = msgpack.unpackb(raw_msg)
-                await self.dispatch_consumers(msg['name'],
-                                              msg['source'],
-                                              msg['args'])
-            except asyncio.CancelledError:
-                break
-            except Exception:
-                log.exception('EventDispatcher.consume(): unexpected-error')
+        async with aclosing(redis.read_stream_by_group(
+            self.redis_client,
+            'events',
+            self._consumer_group,
+            self._consumer_name,
+        )) as agen:
+            async for msg_id, msg_data in agen:
+                try:
+                    await self.dispatch_consumers(
+                        msg_data[b'name'].decode(),
+                        msg_data[b'source'].decode(),
+                        msgpack.unpackb(msg_data[b'args']),
+                    )
+                except asyncio.CancelledError:
+                    raise
+                except Exception:
+                    log.exception('EventDispatcher.consume(): unexpected-error')
 
     async def _subscribe_loop(self) -> None:
-
-        async def _subscribe_impl() -> None:
-            channels = await self.redis_subscriber.subscribe('events.pubsub')
-            async for raw_msg in channels[0].iter():
-                msg = msgpack.unpackb(raw_msg)
-                await self.dispatch_subscribers(msg['name'],
-                                                msg['source'],
-                                                msg['args'])
-
-        while True:
-            try:
-                await redis.execute_with_retries(lambda: _subscribe_impl())
-            except asyncio.CancelledError:
-                break
-            except Exception:
-                log.exception('EventDispatcher.subscribe(): unexpected-error')
+        async with aclosing(redis.read_stream(
+            self.redis_client,
+            'events',
+        )) as agen:
+            async for msg_id, msg_data in agen:
+                try:
+                    await self.dispatch_subscribers(
+                        msg_data[b'name'].decode(),
+                        msg_data[b'source'].decode(),
+                        msgpack.unpackb(msg_data[b'args']),
+                    )
+                except asyncio.CancelledError:
+                    raise
+                except Exception:
+                    log.exception('EventDispatcher.subscribe(): unexpected-error')
 
 
 class EventProducer(aobject):
-    redis_producer: aioredis.Redis
-    producer_lock: asyncio.Lock
-
-    _connector: RedisConnectorFunc
+    redis_client: RedisConnectionInfo
     _log_events: bool
 
-    def __init__(self, connector: RedisConnectorFunc, log_events: bool = False) -> None:
-        self._connector = connector
+    def __init__(
+        self,
+        connector: EtcdRedisConfig,
+        db: int = 0,
+        log_events: bool = False,
+        service_name: str = None,
+    ) -> None:
+        _connector = connector.copy()
+        if service_name:
+            _connector['service_name'] = service_name
+        self.redis_client = redis.get_redis_object(_connector, db=db)
         self._log_events = log_events
 
     async def __ainit__(self) -> None:
-        self.redis_producer = await self._connector()
-        self.producer_lock = asyncio.Lock()
+        pass
 
     async def close(self) -> None:
-        self.redis_producer.close()
-        await self.redis_producer.wait_closed()
+        await self.redis_client.close()
 
     async def produce_event(
         self,
@@ -696,15 +791,12 @@ class EventProducer(aobject):
         *,
         source: str = 'manager',
     ) -> None:
-        raw_msg = msgpack.packb({
-            'name': event.name,
-            'source': source,
-            'args': event.serialize(),
-        })
-        async with self.producer_lock:
-            def _pipe_builder():
-                pipe = self.redis_producer.pipeline()
-                pipe.rpush('events.prodcons', raw_msg)
-                pipe.publish('events.pubsub', raw_msg)
-                return pipe
-            await redis.execute_with_retries(_pipe_builder)
+        raw_event = {
+            b'name': event.name.encode(),
+            b'source': source.encode(),
+            b'args': msgpack.packb(event.serialize()),
+        }
+        await redis.execute(
+            self.redis_client,
+            lambda r: r.xadd('events', raw_event),  # type: ignore # aio-libs/aioredis-py#1182
+        )
