@@ -348,7 +348,11 @@ class RelayHandler(logging.Handler):
             if record is not None and record.exc_info is not None:
                 pickling_support.install(record.exc_info[1])
             pickled_rec = pickle.dumps(record)
-        except (pickle.PickleError, TypeError):
+        except (
+            pickle.PickleError,
+            TypeError,
+            ImportError,  # when "Python is likely to be shutting down"
+        ):
             # We have a pickling error.
             # Change it into a self-created picklable log record with exception info.
             if record is not None:
@@ -456,5 +460,5 @@ class Logger():
             self.log_worker.join()
             self.relay_handler.close()
             ep_url = yarl.URL(self.log_endpoint)
-            if ep_url.scheme.lower() == 'ipc':
-                os.unlink(ep_url.path)
+            if ep_url.scheme.lower() == 'ipc' and (ep_sock := Path(ep_url.path)).exists():
+                ep_sock.unlink()
